/*
	$Id$
	Definitions for the FastCGI application box
	Copyright 2005 Għsiorowski Maciej
	$Revision$
	
	Parts of this code are derived from suexec.c Copyright 1995-1997 The Apache Group, and sbox.c Copyright 1997-1998, Lincoln D. Stein
*/
#ifndef _SBOX_H
#define _SBOX_H

// because sbox-fastcgi is replacement for suexec we have to fool apache
#ifndef _SUEXEC_H
#define _SUEXEC_H
#endif

#define _DEBUG 1

#define TRUE 1
#define FALSE 0

/******************************** GENERAL *************************************/
/*
 * WEB_USER  -- user that the Web server runs as
 */
#ifndef WEB_USER
#define WEB_USER "www-data"
#endif

/*
 * WEB_GROUP  -- group that the Web server runs as
 */
#ifndef WEB_GROUP
#define WEB_GROUP "www-data"
#endif

/*
 * UID_MIN  -- lowest UID that we will suid to
 */
#ifndef UID_MIN
#define UID_MIN 100
#endif

/*
 * UID_MAX  -- highest UID that we will suid to
 */
#ifndef UID_MAX
#define UID_MAX 65535
#endif

/*
 * GID_MIN  -- lowest UID that we will sgid to
 */
#ifndef GID_MIN
#define GID_MIN 100
#endif

/*
 * GID_MAX  -- highest GID that we will sgid to
 */
#ifndef GID_MAX
#define GID_MAX 65535
#endif

#ifndef VHOSTS_ROOT
#define VHOSTS_ROOT "/home/vhosts"
#endif

#ifndef SAFE_PATH
#define SAFE_PATH "/bin:/usr/bin"
#endif


/******************************** LOGGING *************************************/
#ifndef LOG_FILE
#define LOG_FILE "/var/log/sbox/sbox.log" 
#endif

/****************************** PHP SPECIFIC***********************************/
/*
	Additional path for PHP to search for php.ini
	Default fallback is to compiled location. It is set as an env to 
	PHP binary
*/
#ifndef PHPRC_PATH
#define PHPRC_PATH "/www"
#endif

/*
	This is for PHP process manager - number of requests before restarting PHP
	binary
*/
#ifndef PHP_FCGI_MAX_REQUESTS
#define PHP_FCGI_MAX_REQUESTS "100"
#endif

#ifndef PHP_FCGI_CHILDREN
#define PHP_FCGI_CHILDREN "1"
#endif


#endif /* _SBOH_H */
